' Capture
' By: Ted Felix
' QBASIC-ization on 1/20/2003
' Copyright 2003, Ted Felix, All Rights Reserved

' Suggestions to make this more interesting:
' 1> Make increasing difficulty levels and offer a scoring system.
'    More/faster enemies, fewer blocks
' 2> Throw in some special blocks of some kind
'    Traps - Stun/kill enemies or you
'    Immoveable blocks
' 3> Gold coins for bonus points.
' 4> Differing enemy behavior.  Path finding(!)
' 5> Hmm, this is beginning to sound like Escape

' History
' 2/18/03 - Fixed midnight crossing issue.  Still untested, though.

TYPE PlayerState
  X AS INTEGER
  Y AS INTEGER
END TYPE

TYPE EnemyState
  X AS INTEGER
  Y AS INTEGER
  Captured AS INTEGER
END TYPE

DECLARE SUB FindSpace (X AS INTEGER, Y AS INTEGER)
DECLARE SUB Sleep2 (Seconds AS SINGLE)
DECLARE SUB DrawField ()
DECLARE SUB MoveEnemy (oEnemyState() AS EnemyState, oPlayerState AS PlayerState)
DECLARE SUB MovePlayer (oPlayerState AS PlayerState)
DECLARE SUB PushBlocks (X1%, Y1%, XX%, YY%, PX%, PY%)
DECLARE SUB EnemyStuck (EnemyX%, EnemyY%, EnemyStuck%)

' How many seconds before enemy gets to move
CONST EnemyUpdateInterval = .4

' How many blocks on the screen
CONST Blocks = 200

' Boolean algebra constants
CONST true = (1 = 1)
CONST false = (1 = 0)

RANDOMIZE TIMER
DEF FNRNDX = INT(RND * 38) + 2
DEF FNRNDY = INT(RND * 21) + 2

SCREEN 0
WIDTH 40


DO

  CLS

  ' Draw the playing field
  CALL DrawField

  ' Switch to blinking
  COLOR 23, 0

  ' Enemy State
  DIM oEnemyState(2) AS EnemyState

  ' For each enemy
  FOR I = 1 TO 2
    ' Find a place for the enemy
    CALL FindSpace(oEnemyState(I).X, oEnemyState(I).Y)

    ' Draw the enemy
    LOCATE oEnemyState(I).Y, oEnemyState(I).X
    PRINT "*";

    ' Indicate enemy has not been captured
    oEnemyState(I).Captured = 0
  NEXT I

  ' Player State
  DIM oPlayerState AS PlayerState

  ' Find a place for the player
  CALL FindSpace(oPlayerState.X, oPlayerState.Y)

  ' Draw the player
  LOCATE oPlayerState.Y, oPlayerState.X
  PRINT "O";

  ' Wait a bit so the player can see where the enemies are
  Sleep2 (2)

  ' Turn off the blink
  COLOR 7, 0

  ' First time through, enemy moves immediately
  EnemyMoveTime = TIMER

  ' Indicate game is not over yet
  Done% = false
  Win% = false


  ' *** The Game Loop ***

  DO

    ' *** Move the Player ***
    CALL MovePlayer(oPlayerState)

    ' *** Move the Enemies ***

    ' If it's time to move an enemy
    IF TIMER >= EnemyMoveTime THEN

      ' Figure out the next enemy move time
      EnemyMoveTime = TIMER + EnemyUpdateInterval

      ' If the clock is about to turn over (at midnight)
      IF EnemyMoveTime >= 86400 THEN
        EnemyMoveTime = EnemyMoveTime - 86400

        ' Wait until midnight passes
        DO
        LOOP UNTIL TIMER < 100
      END IF

      CALL MoveEnemy(oEnemyState(), oPlayerState)

    END IF

  LOOP WHILE NOT Done%
  
  IF Win% THEN
    CLS
    PRINT "You got 'em!"
  ELSE
    CLS
    PRINT "YOU'RE DEAD!"
  END IF

  ' Wait for 2 seconds
  Sleep2 (2)

  ' Empty the key buffer
  DO
  LOOP WHILE INKEY$ <> ""

  PRINT "Press any key to play again"
  A$ = INPUT$(1)

LOOP

SUB DrawField
  ' Draw the border
  COLOR 4
  LOCATE 1, 1, 0
  S$ = STRING$(40, 177)
  S1$ = CHR$(177)
  PRINT S$;
  FOR I = 1 TO 21
    PRINT S1$; SPACE$(38); S1$;
  NEXT I
  PRINT S$;

  ' Fill the field with blocks
  COLOR 2
  FOR I = 1 TO Blocks
    Y = FNRNDY
    X = FNRNDX
    LOCATE Y, X
    PRINT "";
  NEXT I
END SUB

SUB EnemyStuck (EnemyX%, EnemyY%, Stuck%)
' *** Enemy is stuck ***

' EnemyX = Enemy X
' EnemyY = Enemy Y
' Stuck = Enemy Still Stuck

' Assume stuck
Stuck% = -1

' Try ten random moves
FOR I = 1 TO 10
  EXX% = INT(RND * 3) - 1
  EYY% = INT(RND * 3) - 1
  ESC% = SCREEN(EnemyY% + EYY%, EnemyX% + EXX%)
  ' If the way is clear
  IF ESC% = 32 THEN
    EnemyY% = EnemyY% + EYY%
    EnemyX% = EnemyX% + EXX%
    Stuck% = 0
    ' End the loop
    I = 20
  END IF
NEXT I

' If a valid move was found
IF NOT Stuck% THEN
  EXIT SUB
END IF

' Try all possible moves since the random approach failed
FOR EXX = -1 TO 1
  FOR EYY = -1 TO 1
    ' If this isn't a non-move
    IF NOT (EYY = 0 AND EXX = 0) THEN
      ESC% = SCREEN(EnemyY% + EYY, EnemyX% + EXX)
      ' If there's space to move, do it
      IF ESC% = 32 THEN
        ' Do the move
        EnemyX% = EnemyX% + EXX
        EnemyY% = EnemyY% + EYY
        ' Indicate that the enemy is no longer captured
        Stuck% = 0
        ' End the loops
        EXX = 100
        EYY = 100
      END IF
    END IF
  NEXT EYY
NEXT EXX

END SUB

SUB FindSpace (X AS INTEGER, Y AS INTEGER)
  ' Routine to find a random empty space on the screen.

  DO
    X = FNRNDX
    Y = FNRNDY
  LOOP WHILE SCREEN(Y, X) <> 32

END SUB

SUB MoveEnemy (oEnemyState() AS EnemyState, oPlayerState AS PlayerState)
  SHARED Done%
  SHARED Win%

  STATIC E1

  ' Switch enemies
  IF E1 = 1 THEN E1 = 2 ELSE E1 = 1

  ' Save the enemy's previous position
  PEnemyX = oEnemyState(E1).X
  PEnemyY = oEnemyState(E1).Y

  ' Move the enemy toward the player
  EnemyXDirection = SGN(oPlayerState.X - oEnemyState(E1).X)
  EnemyYDirection = SGN(oPlayerState.Y - oEnemyState(E1).Y)
  oEnemyState(E1).X = oEnemyState(E1).X + EnemyXDirection
  oEnemyState(E1).Y = oEnemyState(E1).Y + EnemyYDirection
  ESC = SCREEN(oEnemyState(E1).Y, oEnemyState(E1).X)
  
  ' Assume the enemy is no longer stuck
  oEnemyState(E1).Captured = 0

  ' If something is in the way
  IF ESC = 219 OR ESC = 177 OR ESC = 42 THEN
    ' Move the enemy back
    oEnemyState(E1).Y = PEnemyY
    oEnemyState(E1).X = PEnemyX
 
    ' Try finding a direction the enemy can go in
    EnemyStuck oEnemyState(E1).X, oEnemyState(E1).Y, oEnemyState(E1).Captured
 
    ' If both enemies have been captured
    IF oEnemyState(1).Captured AND oEnemyState(2).Captured THEN
      Done% = true
      Win% = true
    END IF

  END IF

  ' If the enemy got the player
  IF ESC = 79 THEN
    Done% = true
    Win% = false
  END IF

  ' Move the enemy to its new position
  LOCATE PEnemyY, PEnemyX
  PRINT " ";
  LOCATE oEnemyState(E1).Y, oEnemyState(E1).X
  PRINT "*";

END SUB

SUB MovePlayer (oPlayerState AS PlayerState)
  SHARED Done%
  SHARED Win%

  ' Query the keyboard
  A$ = INKEY$

  ' If it appears to be a special key
  IF LEN(A$) = 2 THEN
    ' Save the player's previous position
    PX% = oPlayerState.X
    PY% = oPlayerState.Y

    ' Assume no motion
    XDirection% = 0
    YDirection% = 0

    R$ = RIGHT$(A$, 1)
    SELECT CASE R$
    CASE "M" ' Right Arrow
      XDirection% = 1
      oPlayerState.X = oPlayerState.X + XDirection%
      IF oPlayerState.X > 39 THEN oPlayerState.X = 39
    CASE "K" ' Left Arrow
      XDirection% = -1
      oPlayerState.X = oPlayerState.X + XDirection%
      IF oPlayerState.X < 2 THEN oPlayerState.X = 2
    CASE "H" ' Up Arrow
      YDirection% = -1
      oPlayerState.Y = oPlayerState.Y + YDirection%
      IF oPlayerState.Y < 2 THEN oPlayerState.Y = 2
    CASE "P" ' Down Arrow
      YDirection% = 1
      oPlayerState.Y = oPlayerState.Y + YDirection%
      IF oPlayerState.Y > 22 THEN oPlayerState.Y = 22
    END SELECT

    ' Check the player's new location
    SC = SCREEN(oPlayerState.Y, oPlayerState.X)

    SELECT CASE SC
    CASE 219  ' If it is a block
      ' Use the block shifting routine
      CALL PushBlocks(oPlayerState.X, oPlayerState.Y, XDirection%, YDirection%, PX%, PY%)
    CASE 42   ' If it is an enemy
      ' Indicate Game Over and the player lost
      Done% = true
      Win% = false
    END SELECT

    ' Clear the old player position
    LOCATE PY%, PX%
    PRINT " ";

    ' Draw the player in the new position
    LOCATE oPlayerState.Y, oPlayerState.X
    PRINT "O";
  END IF

END SUB

SUB PushBlocks (X1%, Y1%, XX%, YY%, PX%, PY%)
  ' *** Push Blocks ***
  ' Inputs:
  '   X1 - Player X position (in/out)
  '   Y1 - Player Y position (in/out)
  '   XX - Player X delta
  '   YY - Player Y delta
  '   PX - Player previous X position
  '   PY - Player previous Y position

  XPush% = X1%
  YPush% = Y1%

  ' Assume a push will happen
  Push = -1

  ' Search in the push direction
  DO
    XPush% = XPush% + XX%
    YPush% = YPush% + YY%
    SPush% = SCREEN(YPush%, XPush%)
    ' If we encounter the border or an enemy
    IF SPush% = 177 OR SPush% = 42 THEN
      ' Move the player back to the original position
      X1% = PX%
      Y1% = PY%
      Push = 0
      ' End the loop
      SPush% = 32
    END IF
  LOOP WHILE SPush% <> 32

  IF Push THEN
    LOCATE YPush%, XPush%
    COLOR 2
    PRINT "";
    COLOR 7
  END IF

END SUB

SUB Sleep2 (Seconds AS SINGLE)
  ' Similar to SLEEP, but cannot be interrupted by keyboard events
 
  ' Compute time in the future when we can continue
  Delay! = TIMER + Seconds

  ' Check for midnight crossing
  IF Delay! >= 86400 THEN
    Delay! = Delay! - 86400

    ' Wait until the midnight crossing is over
    DO
    LOOP UNTIL TIMER < 100
  END IF

  ' Wait for that time to elapse
  DO
  LOOP UNTIL TIMER >= Delay!
END SUB

